/**
 * File for image handling logic.
 */


// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "image.h"
    #include "utility.h"
#endif

#ifndef user_headers_logging
    #define user_headers_logging
    #include "logging.h"
#endif


// Global Variables.
#ifndef LOGGING_CLASS_INIT
    #define LOGGING_CLASS_INIT
    extern Logging logger;
#endif
const Image::PixelStruct Image::pixelBlack = Image::PixelStruct(0);
const Image::PixelStruct Image::pixelWhite = Image::PixelStruct(1);
const Image::PixelStruct Image::pixelRed = Image::PixelStruct(1,0,0);
const Image::PixelStruct Image::pixelGreen = Image::PixelStruct(0,1,0);
const Image::PixelStruct Image::pixelBlue = Image::PixelStruct(0,0,1);
const Image::PixelStruct Image::pixelYellow = Image::PixelStruct(1,1,0);
const Image::PixelStruct Image::pixelPink = Image::PixelStruct(1,0,1);
const Image::PixelStruct Image::pixelLiBlue = Image::PixelStruct(0,1,1);


//region CUDA Methods.

/**
 * Displays various properties of all CUDA devices found on machine.
 **/
void display_cuda_device_properties() {
    // Check available CUDA devices.
    cudaDeviceProp cuda_device_properties;
    int cuda_device_count;

    cudaGetDeviceCount(&cuda_device_count);

    for (int index = 0; index < cuda_device_count; index++) {
        cudaGetDeviceProperties(&cuda_device_properties, index);

        printf("\n\n\n\n");
        printf("==== General Information for CUDA Device %d ====\n", index);
        printf("Name: %s\n", cuda_device_properties.name);
        printf("Compute Capabilities: %d.%d\n", cuda_device_properties.major, cuda_device_properties.minor);
        printf("Clock Rate: %d\n", cuda_device_properties.clockRate);
        printf("\n\n");
        printf("==== Memory Information for CUDA Device %d ====\n", index);
        printf("Total Global Memory: %ld\n", cuda_device_properties.totalGlobalMem);
        printf("Total Constant Memory: %ld\n", cuda_device_properties.totalConstMem);
        printf("Max Memory Pitch: %ld\n", cuda_device_properties.memPitch);
        printf("Texture Alignment: %ld\n", cuda_device_properties.textureAlignment);
        printf("\n\n");
        printf("==== Multiprocessing Information for CUDA Device %d ====\n", index);
        printf("Multiprocessor Count: %d\n", cuda_device_properties.multiProcessorCount);
        printf("Shared Memory Per MP: %ld\n", cuda_device_properties.sharedMemPerBlock);
        printf("Registers Per MP: %d\n", cuda_device_properties.regsPerBlock);
        printf("Threads in Warp: %d\n", cuda_device_properties.warpSize);
        printf("Max Threads Per Block: %d\n", cuda_device_properties.maxThreadsPerBlock);
        printf("Max Thread Dimensions: (%d, %d, %d)\n",
            cuda_device_properties.maxThreadsDim[0],
            cuda_device_properties.maxThreadsDim[1],
            cuda_device_properties.maxThreadsDim[2]
        );
        printf("Max Grid Dimensions: (%d, %d, %d)\n",
            cuda_device_properties.maxGridSize[0],
            cuda_device_properties.maxGridSize[1],
            cuda_device_properties.maxGridSize[2]
        );
        printf("\n\n\n\n");
    }
}


/**
 * Uses convolution to blur image.
 */
__global__ void cuda_convolution_blur(
    unsigned long total_pixels, unsigned long chunk_size,
    int chunk_counter, int image_width, int image_height, int mask_width, int mask_height,
    Image::PixelStruct* pixel_arr_orig, Image::PixelStruct* pixel_arr, Image::PixelStruct* mask_arr
) {
    // Calculate current index.
    unsigned long pixel_index = blockIdx.x * blockDim.x + threadIdx.x;

    // Adjust index for current chunk.
    pixel_index += (chunk_size * chunk_counter);

    // Only proceed if GPU thread corresponds to a pixel in the image. Ignore all extra threads.
    if (pixel_index <= total_pixels) {
        // printf("total_pixels: %ld    pixel_index: %ld\n", total_pixels, pixel_index);

        // Calculate pixel location data.
        int image_row = (pixel_index / image_width) % image_height;
        int image_col = pixel_index % image_width;
        int image_mid_row = image_height / 2;
        int image_mid_col = image_width / 2;

        // Calculate mask location data.
        int mask_mid_row = mask_height / 2;
        int mask_mid_col = mask_width / 2;

        // Calculate pixel locations based on mask values.
        int col_far_left = image_mid_col - mask_mid_col;
        int col_far_right = image_mid_col + mask_mid_col;
        int row_far_top = image_mid_row - mask_mid_row;
        int row_far_bot = image_mid_row + mask_mid_row;

        // // For now, only compute on pixel in center of image.
        // if ((image_row == image_mid_row) && (image_col == image_mid_col)) {
            // printf("\n\n");
            // printf("total_pixels: %ld    image_width: %d    image_height: %d\n", total_pixels, image_width, image_height);
            // printf("Middle pixel found: %ld    image_col: %d    image_row: %d\n", pixel_index, image_col, image_row);
            // printf("image_middle_col: %d     image_middle_row: %d    mask_middle_col: %d    mask_middle_row: %d\n", image_mid_col, image_mid_row, mask_mid_col, mask_mid_row);

            // printf("\n");
            // printf("Associated mask indexes:\n");
            // printf("Center:        [%d,%d]\n", image_mid_col, image_mid_row);
            // printf("CenterLeft:    [%d,%d]\n", col_far_left, image_mid_row);
            // printf("CenterRight:   [%d,%d]\n", col_far_right, image_mid_row);
            // printf("CenterTop:     [%d,%d]\n", image_mid_col, row_far_top);
            // printf("CenterBottom:  [%d,%d]\n", image_mid_col, row_far_bot);
            // printf("\n\n");

            // // Display grid of mask values.
            // for (int row_index = row_far_top; row_index <= row_far_bot; row_index++) {
            //     for (int col_index = col_far_left; col_index <= col_far_right; col_index++) {
            //         printf(" [%d,%d] ", col_index, row_index);
            //     }
            //     printf("\n");
            // }
            // printf("\n\n");

            // // Display grid of mask values.
            // for (int row_index = 0; row_index < image_height; row_index++) {
            //     for (int col_index = 0; col_index < image_width; col_index++) {
            //         int a = col_index;
            //         int b = row_index;
            //         int c = a + (b * image_width);
            //         printf(" [%d,%d] (%d) ", a, b, c);
            //     }
            //     printf("\n");
            // }
            // printf("\n\n");

            // // Display grid of mask values.
            // for (int row_index = row_far_top; row_index <= row_far_bot; row_index++) {
            //     for (int col_index = col_far_left; col_index <= col_far_right; col_index++) {
            //         int a = col_index;
            //         int b = row_index;
            //         int c = a + (b * image_width);
            //         printf("  [%d,%d] (%d) ", a, b, c);
            //     }
            //     printf("\n");
            // }
            // printf("\n\n");

            // pixel_arr[pixel_index].r = 1.f;
            // pixel_arr[pixel_index].g = 0.f;
            // pixel_arr[pixel_index].b = 0.f;

            // Overlay mask onto current pixel, and iterate through all associated pixels.
            float convolution_total_r = 0.f;
            float convolution_total_g = 0.f;
            float convolution_total_b = 0.f;
            float convolution_weight_total_r = 0.f;
            float convolution_weight_total_g = 0.f;
            float convolution_weight_total_b = 0.f;
            int convolution_counter = 0;
            int mask_index = 0;
            for (int row_index = row_far_top; row_index <= row_far_bot; row_index++) {
                for (int col_index = col_far_left; col_index <= col_far_right; col_index++) {

                    // Double check that mask index is within image bounds. Ignore otherwise.
                    if ( ( (col_index >= 0) && (col_index < image_width) ) && ( (row_index >= 0) && (row_index < image_height) ) ) {
                        int arr_index = col_index + (row_index * image_width);

                    // //     // int a = col_index;
                    // //     // int b = row_index;
                    // //     // int c = a + (b * image_width);
                    // //     // printf("  [%d,%d] (%d) (%f,%f,%f)", a, b, c, pixel_arr[mask_index].r, pixel_arr[mask_index].g, pixel_arr[mask_index].b);

                        convolution_total_r += pixel_arr_orig[arr_index].r * mask_arr[mask_index].r;
                        convolution_total_g += pixel_arr_orig[arr_index].g * mask_arr[mask_index].g;
                        convolution_total_b += pixel_arr_orig[arr_index].b * mask_arr[mask_index].b;
                        convolution_counter += 1;


                        if ( (col_index == image_mid_col) && (row_index == image_mid_row) ) {

                            // float mask_inverse = 1.f - mask_arr[mask_index].r;
                            // float updated_r = pixel_arr_orig[arr_index].r * mask_arr[mask_index].r;
                            // float updated_inv_r = pixel_arr_orig[pixel_index].r * mask_inverse;
                            // printf("[%d,%d]    orig_r: %f    mask_weight: %f    mask_inv: %f\n", col_index, row_index, pixel_arr_orig[arr_index].r, mask_arr[mask_index].r, mask_inverse);
                            // printf("updated_r: %f    updated_inv_r: %f\n", updated_r, updated_inv_r);
                            // updated_r = updated_r + updated_inv_r;
                            // // printf("total_pixels: %ld    pixel_index: %ld    mask_index: %d\n", total_pixels, pixel_index, mask_index);
                            // printf("updated_r: %f\n", updated_r);
                            // // printf("[%d,%d]    orig_r: %f    mask_weight: %f    mask_inv: %f    new_r: %f\n", col_index, row_index, pixel_arr_orig[arr_index].r, mask_arr[mask_index].r, mask_inverse, updated_r);



                        }

                        // Calculate updated red value.
                        float mask_inverse = 1.f - mask_arr[mask_index].r;
                        float updated_pix = pixel_arr_orig[arr_index].r * mask_arr[mask_index].r;
                        float updated_inv_pix = pixel_arr_orig[pixel_index].r * mask_inverse;
                        convolution_total_r = convolution_total_r + updated_pix + updated_inv_pix;
                        convolution_weight_total_r += mask_arr[mask_index].r;

                        // Calculate updated green value.
                        mask_inverse = 1.f - mask_arr[mask_index].g;
                        updated_pix = pixel_arr_orig[arr_index].g * mask_arr[mask_index].g;
                        updated_inv_pix = pixel_arr_orig[pixel_index].g * mask_inverse;
                        convolution_total_g = convolution_total_g + updated_pix + updated_inv_pix;
                        convolution_weight_total_g += mask_arr[mask_index].g;

                        // Calculate updated blue value.
                        mask_inverse = 1.f - mask_arr[mask_index].b;
                        updated_pix = pixel_arr_orig[arr_index].b * mask_arr[mask_index].b;
                        updated_inv_pix = pixel_arr_orig[pixel_index].b * mask_inverse;
                        convolution_total_b = convolution_total_b + updated_pix + updated_inv_pix;
                        convolution_weight_total_b += mask_arr[mask_index].b;


                    //     // float red = pixel_arr_orig[arr_index].r * mask_arr[mask_index].r;
                    //     // float green = pixel_arr_orig[arr_index].g * mask_arr[mask_index].g;
                    //     // float blue = pixel_arr_orig[arr_index].b * mask_arr[mask_index].b;
                    //     // pixel_arr[arr_index].r = red;
                    //     // pixel_arr[arr_index].g = green;
                    //     // pixel_arr[arr_index].b = blue;
                    }
                    mask_index += 1;
                }
                // printf("\n");
            }

            // Average out values and apply to pixel index.
            if (convolution_total_r > 0.f) {
                pixel_arr[pixel_index].r = (convolution_total_r / convolution_counter);
                // pixel_arr[pixel_index].r = (1.f / convolution_weight_total_r) * convolution_total_r;
            } else {
                pixel_arr[pixel_index].r = 0.f;
            }
            if (convolution_total_g > 0.f) {
                pixel_arr[pixel_index].g = (convolution_total_g / convolution_counter);
                // pixel_arr[pixel_index].g = (1.f / convolution_weight_total_g) * convolution_total_g;
            } else {
                pixel_arr[pixel_index].g = 0.f;
            }
            if (convolution_total_b > 0.f) {
                pixel_arr[pixel_index].b = (convolution_total_b / convolution_counter);
                // pixel_arr[pixel_index].b = (1 / convolution_weight_total_b) *  convolution_total_b;
            } else {
                pixel_arr[pixel_index].b = 0.f;
            }

            // printf("\n\n");
        // }
    }
}

//endregion CUDA Methods.


//region Constructors.

/**
 * Base constructor.
 */
Image::Image() {
    logger.debug("Image::Image Constructor():");
    logger.debug("");

    // Currently test logic.

    // Essentially creates a 10x10 pixel test image.
    input_path = "../input/pbm/400x300/hallway_00.pbm";
    output_path = "../output/test_00.pbm";
    width = 10;
    height = 10;
    bit_depth = 255;
    PixelStruct c = pixelBlack; // Default color. Can select based on below preset colors.

    // Create and populate image.
    pixel_arr = new PixelStruct[width * height];
    for (int index = 0; index < (width * height); index++) {
        pixel_arr[index] = c;
    }
}


/**
 * Constructor that takes input path to image file.
 * Output path is determined based on parsing input path.
 */
Image::Image(std::string user_path) {
    logger.debug("Image::Image Constructor():");
    logger.debug("    user_path: " + user_path);
    logger.debug("");

    logger.info("Constructing image class.");

    // Re-validate path for saving to class.
    path_validator_struct* return_struct = validate_path(user_path);
    int path_type = return_struct->file_type;
    if (return_struct->err_code) {
        logger.error("Error verifying file type: " + return_struct->err_code.message());
        return;
    }
    free_path_validator_struct(return_struct);
    if (path_type == TYPE_FILE) {
        input_path = user_path;
        calculate_paths(user_path);
        import_image_file();
    } else {
        logger.error("Provided path is not an image file. Class initialization failed.");
        return;
    }

    logger.info("Image class created.");
}


/**
 * Deconstructor.
 */
Image::~Image() {
    logger.debug("Deconstructing image class");

    // Deallocate memory objects.
    delete[] pixel_arr;
}

//endregion Constructors.


//region Private Methods.

/**
 * Calculates class input_path and output_path values.
 *
 * :param image_path: Path to an image on the system.
 */
void Image::calculate_paths(std::string image_path) {
    logger.debug("Image::calculate_paths():");
    logger.debug("    image_path: " + image_path);
    logger.debug("");

    // // Calculate output path location.
    output_path = "../output/";
    std::string base_filename = get_file_name();;

    // Grab directory substrings and generate full output path.
    size_t pos = 0;
    std::string token;
    while ((pos = image_path.find("/")) != std::string::npos) {

        // Pull current substring value from input path.
        token = image_path.substr(0, image_path.find("/"));
        image_path.erase(0, pos + 1);

        // Check if directory value is value we care about.
        if ( (token != ".") && (token != "..") && (token != "input") ) {
            output_path += token + "/";
        }

        // Check if directory exists.
        struct stat info;
        if ( (stat(output_path.c_str(), &info) != 0) || (! info.st_mode & S_IFDIR)) {
            // Directory does not exist.
            mkdir(output_path.c_str(), 0777);
        }

    }

    output_path += base_filename;
}


/**
 * Imports image file data, using class variable data.
 */
void Image::import_image_file() {
    logger.debug("Image::import_image_file():");
    logger.debug("");

    logger.info("Attempting to import image data...");
    logger.info("Data from: \"" + input_path + "\"");

    // Open input file.
    std::ifstream input_file;
    input_file.open(input_path, std::ios::binary | std::ios::in);

    if (! input_file.good()) {
        // Failed to open file.
        logger.error("Failed to open file \"" + input_path + "\"");
    } else {
        // File opened. Grab data.
        std::string header_format;
        input_file >> header_format;

        // Check image data format type.
        if (strcmp(header_format.c_str(), "P6") == 0) {
            // P6 indicates standard "binary RGB format".
            logger.debug("Image is binary RGB format.");

        } else if (strcmp(header_format.c_str(), "P5") == 0) {
            // P5 indicates standard "binary greyscale format".
            logger.debug("Image is binary greyscale format.");

        } else {
            // throw("Can't read input file");
            logger.error("");
            logger.error("Image: \"" + input_path + "\"");
            logger.error("Unhandled image format type: " + header_format);
            return;
        }

        // Save image values to class.
        input_file >> width >> height >> bit_depth;

        // Create and populate image.
        pixel_arr = new PixelStruct[width * height];

        // Skip empty lines in necessary until we get to the binary data.
        input_file.ignore(256, '\n');

        // Read pixels based on image format.
        if (strcmp(header_format.c_str(), "P6") == 0) {
            // Binary RBG format.
            // Read each pixel one by one and convert RGB channel bytes to floats.
            unsigned char pixel[3];
            for (int index = 0; index < (width * height); ++index) {
                input_file.read(reinterpret_cast<char *>(pixel), 3);
                pixel_arr[index].r = pixel[0] / 255.f;
                pixel_arr[index].g = pixel[1] / 255.f;
                pixel_arr[index].b = pixel[2] / 255.f;
            }
        } else if (strcmp(header_format.c_str(), "P5") == 0) {
            // Binary RBG format.
            // Read each pixel one by one and convert bytes to floats.
            unsigned char* pixel;
            pixel = (unsigned char*)calloc(1, sizeof(unsigned char));
            for (int index = 0; index < (width * height); ++index) {
                input_file.read(reinterpret_cast<char *>(pixel), 1);
                pixel_arr[index].r = *pixel / 255.f;
                pixel_arr[index].g = *pixel / 255.f;
                pixel_arr[index].b = *pixel / 255.f;
            }
            free(pixel);
        }
    }

    input_file.close();
}

//endregion Private Methods.


//region Public Methods.

/**
 * Displays general properties of image held in class.
 */
void Image::display_properties() {
    logger.debug("Image::display_properties():");
    logger.debug("");

    logger.info("");
    logger.info("input file path: " + input_path);
    logger.info("output file path: " + output_path);
    logger.info("width: " + std::to_string(width));
    logger.info("height: " + std::to_string(height));
    logger.info("bit depth: " + std::to_string(bit_depth));
    logger.info("");
}


/**
 * Saves class image data to output file.
 */
void Image::save() {
    logger.debug("Image::save():");
    logger.debug("");

    logger.debug("Attempting to save image to \"" + output_path + "\".");

    // Open output file stream.
    std::ofstream output_file;

    // Attempt to save file data.
    try {
        // Attempt to open file.
        output_file.open(output_path, std::ios::binary | std::ios::out);

        // Check if opened successfully.
        if (output_file.fail()) {
            // Failed to open.
            logger.error("Failed to open file \"" + output_path + "\"");
        } else {
            // Opened successfully.

            // Write image "header" data.
            output_file << "P6\n" << width << " " << height << "\n255\n";

            // Write image pixel data.
            unsigned char r, g, b;
            for (int index = 0; index < (width * height); index++) {
                r = static_cast<float>(std::min(1.f, pixel_arr[index].r) * 255);
                g = static_cast<float>(std::min(1.f, pixel_arr[index].g) * 255);
                b = static_cast<float>(std::min(1.f, pixel_arr[index].b) * 255);
                output_file << r << g << b;
            }

        }

        // Close file stream.
        output_file.close();
        logger.info("Image saved.");

    } catch (const char* err) {
        logger.error("Failed to save image: " + std::string(err));
        output_file.close();
    }
}


/**
 * Gets current file name, using original inpute path value.
 */
std::string Image::get_file_name() {
    logger.debug("Image::get_file_name():");
    logger.debug("");

    return input_path.substr(input_path.find_last_of("/\\") + 1);
}


/**
 * Changes output file name to provided value.
 */
void Image::set_file_name(std::string file_name) {
    logger.debug("Image::compute():");
    logger.debug("    file_name: " + file_name);
    logger.debug("");

    std::string new_path = output_path.substr(0, output_path.find_last_of("/\\") + 1);
    output_path = new_path + file_name;
}


/**
 * Method to actually do work on image.
 */
void Image::compute() {
    logger.debug("Image::compute():");
    logger.debug("");

    // Get required CUDA device data.
    // For now, lazily gets first CUDA device and assumes it's the desired device to use.
    // TODO: Handle multiple CUDA devices.
    cudaDeviceProp cuda_device_properties;
    cudaGetDeviceProperties(&cuda_device_properties, 0);
    int cuda_device_max_blocks = cuda_device_properties.maxThreadsDim[0];
    int cuda_device_max_threads = cuda_device_properties.maxThreadsPerBlock;
    unsigned long cuda_chunk_size = cuda_device_max_blocks * cuda_device_max_threads;

    // Calculate some image data.
    int pixel_struct_size = sizeof(PixelStruct);
    unsigned long image_total_pixels = width * height;
    unsigned long pixel_struct_byte_count = image_total_pixels * pixel_struct_size;

    // We handle image in chunks, in case we're dealing with large images with more pixels than CUDA device blocks/threads.
    int total_image_chunks = (image_total_pixels / cuda_chunk_size) + 1;

    // Generate convolution mask dimensions. Should be either 10% of image size or 3 pixels large, whichever is greater.
    int mask_width = width / 10;
    int mask_height = height / 10;
    if ((mask_width % 2) == 0) {
        mask_width -= 1;
    }
    if ((mask_height % 2) == 0) {
        mask_height -= 1;
    }
    if (mask_width < 3) {
        mask_width = 3;
    }
    if (mask_height < 3) {
        mask_height = 3;
    }
    int mask_middle_col = mask_width / 2;
    int mask_middle_row = mask_height / 2;

    // Calculate additional mask initialization values.
    // Found via much struggling and trial + error.
    float mask_col_adj = 1.f / (mask_middle_col + 1);
    float mask_row_adj = 1.f / (mask_middle_row + 1);
    // printf("mask_middle_col: %d    mask_middle_row: %d\n", mask_middle_col, mask_middle_row);
    // printf("mask_col_adj: %f    mask_row_adj: %f\n", mask_col_adj, mask_row_adj);
    unsigned long mask_total_pixels = mask_width * mask_height;
    unsigned long mask_struct_byte_count = mask_total_pixels * pixel_struct_size;

    // Initialize mask and set weights.
    // printf("\n\n");
    logger.info("Mask size: " + std::to_string(mask_width) + "x" + std::to_string(mask_height));
    PixelStruct* mask_arr = new PixelStruct[mask_total_pixels];
    for (int mask_row_index = 0; mask_row_index < mask_height; mask_row_index++) {
        for (int mask_col_index = 0; mask_col_index < mask_width; mask_col_index++) {
            // Get current overall mask index, based on row and col values.
            int mask_index = mask_col_index + (mask_row_index * mask_width);

            // Calculate individual weights based on current x and y axis location of mask index, using distance from center point.
            float mask_col_weight = (fabs(mask_middle_col - fabs(mask_middle_col - mask_col_index)) + 1) * mask_col_adj;
            float mask_row_weight = (fabs(mask_middle_row - fabs(mask_middle_row - mask_row_index)) + 1) * mask_row_adj;

            // Calculate overall index weight, based on combination of individual dimension weights.
            float index_weight = (mask_col_weight * mask_row_weight);

            // printf("  [%d, %d] (%d) (%f,%f)  ", mask_col_index, mask_row_index, mask_index, mask_col_weight, mask_row_weight);
            // printf("  [%d, %d] (%d) (%f)  ", mask_col_index, mask_row_index, mask_index, index_weight);

            // Finally, set mask index values.
            mask_arr[mask_index].r = index_weight;
            mask_arr[mask_index].g = index_weight;
            mask_arr[mask_index].b = index_weight;
            // mask_arr[mask_index].r = 1.f;
            // mask_arr[mask_index].g = 1.f;
            // mask_arr[mask_index].b = 1.f;
        }
        // printf("\n");
    }
    // printf("\n\n");

    // Allocate memory on GPU.
    PixelStruct* gpu_pixel_arr;
    PixelStruct* gpu_pixel_arr_orig;
    PixelStruct* gpu_mask_arr;
    cudaMalloc((void**) &gpu_pixel_arr, pixel_struct_byte_count);
    cudaMalloc((void**) &gpu_pixel_arr_orig, pixel_struct_byte_count);
    cudaMalloc((void**) &gpu_mask_arr, mask_struct_byte_count);
    cudaMemcpy(gpu_pixel_arr, pixel_arr, pixel_struct_byte_count, cudaMemcpyHostToDevice);
    cudaMemcpy(gpu_pixel_arr_orig, pixel_arr, pixel_struct_byte_count, cudaMemcpyHostToDevice);
    cudaMemcpy(gpu_mask_arr, mask_arr, mask_struct_byte_count, cudaMemcpyHostToDevice);

    // Process image by iterating over all chunks.
    for (int chunk_index = 0; chunk_index < total_image_chunks; chunk_index++) {
        logger.info("Iterating over image chunk " + std::to_string(chunk_index));
        logger.info("Accounts for pixels " + std::to_string(cuda_chunk_size * chunk_index) + " through " + std::to_string((cuda_chunk_size * chunk_index) + cuda_chunk_size) );

        // Run GPU logic to use convolution to create blur effect.
        cuda_convolution_blur<<<cuda_device_max_blocks, cuda_device_max_threads>>>(image_total_pixels, cuda_chunk_size, chunk_index, width, height, mask_width, mask_height, gpu_pixel_arr_orig, gpu_pixel_arr, gpu_mask_arr);

        // Synchronize all GPU computations before iterating over next chunk.
        cudaDeviceSynchronize();
    }

    // Retrieve results.
    cudaMemcpy(pixel_arr, gpu_pixel_arr, pixel_struct_byte_count, cudaMemcpyDeviceToHost);
    cudaFree(gpu_pixel_arr);
    cudaFree(gpu_pixel_arr_orig);
    cudaFree(gpu_mask_arr);
    delete[] mask_arr;

    logger.debug("Image::compute() finished.");
    logger.debug("");
}

//endregion Public Methods.
